@@ -33,3 +33,8 @@
 #define	KMFREE(v, z)		kmem_free(v, z)
 
 extern	caddr_t			pfil_nd;
+
+#if defined(atomic_add_long) && (SOLARIS2 < 7)
+# undef atomic_add_long
+# define atomic_add_long(x,y)	atomic_add_32((uint32_t *)x, y)
+#endif
