@@ -9,6 +9,7 @@
 
 #include <sys/conf.h>
 #include <sys/debug.h>
+#include <sys/atomic.h>
 #include <sys/ethernet.h>
 #include <sys/stream.h>
 #include <sys/errno.h>
@@ -402,9 +403,7 @@ static int pfilmodclose(queue_t *q, int flag, cred_t *crp)
 
 	qprocsoff(q);
 
-	WRITE_ENTER(&pfil_rw);
 	qif_delete(q->q_ptr, q);
-	RW_EXIT(&pfil_rw);
 	return 0;
 }
 
@@ -414,8 +413,11 @@ static int pfilmodclose(queue_t *q, int flag, cred_t *crp)
 
 /* ------------------------------------------------------------------------ */
 /* Function:    pfil_precheck                                               */
-/* Returns:     int - < 0 is error in this function, 0 == pass packet, else */
-/*                    (> 0) indicates passing prohibited                    */ 
+/* Returns:     int - < 0 pass packet because it's not a type subject to    */
+/*                    firewall rules (i.e. internal STREAMS messages),      */
+/*                    0 == pass packet, else > 0 indicates passing          */
+/*                    prohibited (possibly due to an error occuring in      */
+/*                    this function.)                                       */
 /* Parameters:  q(I)   - pointer to STREAMS queue                           */
 /*              mp(I)  - pointer to STREAMS message                         */
 /*              qif(I) - pointer to per-queue interface information         */
@@ -433,12 +435,12 @@ int pfil_precheck(queue_t *q, mblk_t **mp, int flags, qif_t *qif)
 {
 	register struct ip *ip;
 	size_t hlen, len, off, mlen, iphlen, plen;
-	int err, out, sap, realigned = 0;
 	packet_filter_hook_t *pfh;
 	qpktinfo_t qpkt, *qpi;
 	struct pfil_head *ph;
 	mblk_t *m, *mt = *mp;
-	u_char *bp, *s;
+	int err, out, sap;
+	u_char *bp;
 #if SOLARIS2 >= 8
 	ip6_t *ip6;
 #endif
@@ -468,7 +470,7 @@ int pfil_precheck(queue_t *q, mblk_t **mp, int flags, qif_t *qif)
 	 */
 	out = (flags & PFIL_OUT) ? 1 : 0;
 	off = (out) ? qpi->qpi_hl : 0;
-tryagain:
+
 	ip = NULL;
 	m = NULL;
 
@@ -543,54 +545,120 @@ tryagain:
 
 	}
 
-	/*
-	 * If there is more than one copy of this message traversing the
-	 * STREAMS stack (ie packet is being used for snoop data) then make a
-	 * copy of it for our use so we become the sole owner of the new
-	 * message and do a freemsg() on the one passed in as we're no longer
-	 * using it or passing it up.
-	 */
-	if ((pfil_delayed_copy == 0) && (m->b_datap->db_ref > 1)) {
-		mblk_t *new;
-
-forced_copy:
-		new = copymsg(m);
-		if (new == NULL) {
-			atomic_add_long(&qif->qf_copyfail, 1);
-			return -3;
-		}
-		atomic_add_long(&qif->qf_copy, 1);
-
-		if (mt != m)
-			mt->b_cont = new;
-		else {
-			*mp = new;
-			mt = new;
-		}
-		freemsg(m);
-		m = new;
-	}
-
-	ip = (struct ip *)(m->b_rptr + off);
-
 	/*
 	 * We might have a 1st data block which is really M_PROTO, i.e. it is
 	 * only big enough for the link layer header
 	 */
-	while ((u_char *)ip >= m->b_wptr) {
-		len = (u_char *)ip - m->b_wptr;
+	while ((len = m->b_wptr - m->b_rptr) <= off) {
+		off -= len;
 		m = m->b_cont;
 		if (m == NULL) {
 			atomic_add_long(&qif->qf_nodata, 1);
 			return -4;	/* not enough data for IP */
 		}
-		ip = (struct ip *)(m->b_rptr + len);
 	}
-	off = (u_char *)ip - m->b_rptr;
-	mlen = msgdsize(m) - off;
+
+	ip = (struct ip *)(m->b_rptr + off);
+	len = m->b_wptr - m->b_rptr - off;
+	mlen = msgdsize(m);
 	sap = qif->qf_sap;
 	if (mlen == 0)
-		mlen = mt->b_wptr - mt->b_rptr;
+		mlen = m->b_wptr - m->b_rptr;
+	mlen -= off;
+
+#ifdef PFILDEBUG
+	/*LINTED: E_CONSTANT_CONDITION*/
+	PRINT(10,(CE_CONT,
+		  "!IP Filter[%s]: out %d len %ld/%ld sap %d ip %p b_rptr %p off %ld m %p/%d/%ld/%p mt %p/%d/%ld/%p\n",
+		  qif->qf_name, out, len, mlen, sap,
+		  (void *)ip, (void *)m->b_rptr, off, 
+		  (void *)m, MTYPE(m), MLEN(m), (void *)m->b_cont,
+		  (void *)mt, MTYPE(mt), MLEN(mt), (void *)mt->b_cont));
+#endif
+
+	/*
+	 * If there is more than one copy of this message traversing the
+	 * STREAMS stack (ie the packet is being used for snoop data), the
+	 * IP header isn't on a 32bit aligned address, or the IP header
+	 * isn't contain within a single block, then make a copy which
+	 * meets our requirements and do a freemsg on the one passed in
+	 * since we're no longer using it or passing it up.
+	 */
+
+	if ((pfil_delayed_copy == 0 && m->b_datap->db_ref > 1)
+	    || ((u_int)ip & 0x3) || len < sizeof(*ip)
+	    || (sap != ETHERTYPE_IP
+#if SOLARIS2 >= 8
+		&& sap != IP6_DL_SAP
+#endif
+	        )) {
+		mblk_t *b;
+		mblk_t *nm;
+		mblk_t *nmt;
+		mblk_t *previous_nm;
+
+forced_copy:
+		nmt = NULL;
+		previous_nm = NULL;
+
+		/*
+		 * Duplicate the message block descriptors up to (and
+		 * including if the offset is non-zero) the block where
+		 * IP begins.
+		 */
+		for (b = mt; b != m || off; b = b->b_cont) {
+			nm = dupb(b);
+			if (nm == NULL) {
+				atomic_add_long(&qif->qf_copyfail, 1);
+				if (nmt)
+					freemsg(nmt);
+				return ENOBUFS;
+			}
+
+			nm->b_cont = NULL;
+			if (nmt)
+				linkb(previous_nm, nm);
+			else
+				nmt = nm;
+			previous_nm = nm;
+
+			/*
+			 * Set the length so the block only contains what
+			 * appears before IP.
+			 */
+			if (b == m) {
+				nm->b_wptr = nm->b_rptr + off;
+				break;
+			}
+		}
+
+		m->b_rptr += off;
+		nm = msgpullup(m, -1);
+		m->b_rptr -= off;
+
+		if (nm == NULL) {
+			atomic_add_long(&qif->qf_copyfail, 1);
+			if (nmt)
+				freemsg(nmt);
+			return ENOBUFS;
+		}
+
+		if (nmt)
+			linkb(previous_nm, nm);
+		else
+			nmt = nm;
+
+		freemsg(mt);
+
+		*mp = nmt;
+		mt = nmt;
+		m = nm;
+
+		ip = (struct ip *)m->b_rptr;
+		len = m->b_wptr - m->b_rptr;
+		mlen = len;
+		off = 0;
+	}
 
 	if (sap == ETHERTYPE_IP) {
 		u_short tlen;
@@ -602,7 +670,6 @@ forced_copy:
 		((char *)&tlen)[1] = ((char *)&ip->ip_len)[1];
 		plen = ntohs(tlen);
 
-		sap = 0;
 		ph = &pfh_inet4;
 	}
 #if SOLARIS2 >= 8
@@ -618,119 +685,19 @@ forced_copy:
 		plen = ntohs(tlen);
 		if (plen == 0)
 			return EMSGSIZE;	/* Jumbo gram */
-		plen += sizeof(*ip6);
 
-		sap = IP6_DL_SAP;
 		ph = &pfh_inet6;
 	}
 #endif 
 	else {
-		hlen = 0;
 		sap = -1;
 	}
 
-	len = m->b_wptr - m->b_rptr - off;
-#ifdef PFILDEBUG
-	/*LINTED: E_CONSTANT_CONDITION*/
-	PRINT(10,(CE_CONT,
-		  "!IP Filter[%s]: out %d len %ld/%ld sap %d ip %p b_rptr %p off %ld m %p/%d/%ld/%p mt %p/%d/%ld/%p\n",
-		  qif->qf_name, out, len, mlen, sap,
-		  (void *)ip, (void *)m->b_rptr, off, 
-		  (void *)m, MTYPE(m), MLEN(m), (void *)m->b_cont,
-		  (void *)mt, MTYPE(mt), MLEN(mt), (void *)mt->b_cont));
-#endif
-
-	/*
-	 * Ok, the IP header isn't on a 32bit aligned address so fix this.
-	 */
-	if (((u_int)ip & 0x3) || (len < sizeof(*ip)) || (sap == -1)) {
-		mblk_t *m2, *m1;
-		int off2;
-
-		if (m->b_datap->db_ref > 1)
-			goto forced_copy;
-		/*
-		 * If we have already tried to realign the IP header and we
-		 * are back here, then the attempt has failed, so stop now
-		 * rather than try again (could keep on retrying with no
-		 * benefit.)
-		 */
-		if (realigned) {
-			atomic_add_long(&qif->qf_drop, 1);
-			return EINVAL;
-		}
-		realigned = 1;
-
-		len = msgdsize(m);
-		if (len < sizeof(*ip)) {
-			atomic_add_long(&qif->qf_bad, 1);
-			return EINVAL;
-		}
-
-		/*
-		 * XXX - Now I understand how pullupmsg() & STREAMS messages
-		 * work better, this can possibly be junked in favour of using
-		 * pullupmsg() which will preserve all the dblk bits correctly,
-		 * as is done in fr_pullup in the ipf code.
-		 */
-
-		/*
-		 * Junk using pullupmsg()
-		 */
-		off2 = (u_int)ip & 0x3;
-		if (off2)
-			off2 = 4 - off2;
-
-		m2 = allocb(len + off2, BPRI_HI);
-		if (m2 == NULL) {
-			atomic_add_long(&qif->qf_drop, 1);
-			return ENOBUFS;
-		}
-
-		MTYPE(m2) = M_DATA;
-		if (m->b_rptr != (u_char *)ip)
-			m2->b_rptr += off2;
-		m2->b_wptr = m2->b_rptr + len;
-		m1 = m;
-		s = (u_char *)m->b_rptr;
-		for (bp = m2->b_rptr; m1 && (bp < m2->b_wptr); bp += len) {
-			len = MIN(m1->b_wptr - s, m2->b_wptr - bp);
-			bcopy(s, bp, len);
-			m1 = m1->b_cont;
-			if (m1 != NULL)
-				s = m1->b_rptr;
-		}
-
-		if ((mt != m) && (mt->b_cont == m) && (off == 0)) {
-			/*
-			 * check if the buffer we're changing is chained in-
-			 * between other buffers and unlink/relink as required.
-			 */
-			(void) unlinkb(mt);	/* should return 'm' */
-			m1 = unlinkb(m);
-			if (m1 != NULL)
-				linkb(m2, m1);
-			freemsg(m);
-			linkb(mt, m2);
-		} else {
-			if (m == mt) {
-				m1 = unlinkb(mt);
-				if (m1)
-					linkb(m2, m1);
-			}
-			freemsg(mt);
-			*mp = m2;
-			mt = m2;
-		}
-
-		off = 0;
-		goto tryagain;
-	}
-
-	if (((sap == 0) && (ip->ip_v != IPVERSION))
+	if (((sap == ETHERTYPE_IP) && (ip->ip_v != IPVERSION))
 #if SOLARIS2 >= 8
 	    || ((sap == IP6_DL_SAP) && (((ip6->ip6_vfc) & 0xf0) != 0x60))
 #endif
+	    || sap == -1
 	   ) {
 		atomic_add_long(&qif->qf_notip, 1);
 #ifdef PFILDEBUG
@@ -739,30 +706,7 @@ forced_copy:
 		return EINVAL;
 	}
 
-	/*
-	 * The code in IPFilter assumes that both the ip_off and ip_len
-	 * fields are in host byte order, so convert them here to fulfill
-	 * that expectation.
-	 *
-	 * If the target compile host is non-SPARC, assume it is a little
-	 * endian machine, requiring the conversion of offset/length fields
-	 * to both be host byte ordered.
-	 */
-#ifndef sparc
-# if SOLARIS2 >= 8
-	if (sap == IP6_DL_SAP) {
-		ip6->ip6_plen = plen;
-	} else {
-# endif
-		__ipoff = (u_short)ip->ip_off;
-		ip->ip_len = plen;
-		ip->ip_off = ntohs(__ipoff);
-# if SOLARIS2 >= 8
-	}
-# endif
-#endif
-
-	if (sap == 0)
+	if (sap == ETHERTYPE_IP)
 		iphlen = ip->ip_hl << 2;
 #if SOLARIS2 >= 8
 	else if (sap == IP6_DL_SAP)
@@ -771,27 +715,13 @@ forced_copy:
 
 	if ((
 #if SOLARIS2 >= 8
-	     (sap == IP6_DL_SAP) && (mlen < iphlen + plen)) ||
-	    ((sap == 0) &&
+	     (sap == IP6_DL_SAP) && (mlen < plen)) ||
+	    ((sap == ETHERTYPE_IP) &&
 #endif 
 	     ((iphlen < hlen) || (iphlen > plen) || (mlen < plen)))) {
 		/*
 		 * Bad IP packet or not enough data/data length mismatches
 		 */
-#ifndef sparc
-# if SOLARIS2 >= 8
-		if (sap == IP6_DL_SAP) {
-			ip6->ip6_plen = htons(plen);
-		} else {
-# endif
-			__ipoff = (u_short)ip->ip_off;
-
-			ip->ip_len = htons(plen);
-			ip->ip_off = htons(__ipoff);
-# if SOLARIS2 >= 8
-		}
-# endif
-#endif
 		atomic_add_long(&qif->qf_bad, 1);
 		return EINVAL;
 	}
@@ -805,21 +735,44 @@ forced_copy:
 			goto forced_copy;
 		if (!pullupmsg(m, (int)iphlen + off)) {
 			atomic_add_long(&qif->qf_nodata, 1);
-			return -5;
+			return ENOBUFS;
 		}
 		ip = (struct ip *)ALIGN32(m->b_rptr + off);
 	}
 
-	if ((len > plen) && (off == 0))
-		m->b_wptr -= len - plen;
+	/*
+	 * Discard any excess data.
+	 */
+	if (len > plen)
+		m->b_wptr = m->b_rptr + off + plen;
+
+	/*
+	 * The code in IPFilter assumes that both the ip_off and ip_len
+	 * fields are in host byte order, so convert them here to fulfill
+	 * that expectation.
+	 *
+	 * If the target compile host is non-SPARC, assume it is a little
+	 * endian machine, requiring the conversion of offset/length fields
+	 * to both be host byte ordered.
+	 */
+#ifndef sparc
+	if (sap == ETHERTYPE_IP) {
+		__ipoff = (u_short)ip->ip_off;
+		ip->ip_len = plen;
+		ip->ip_off = ntohs(__ipoff);
+	}
+#endif
 
 	qpi->qpi_m = m;
 	qpi->qpi_off = off;
 	qpi->qpi_data = ip;
 
+	if (qif->qf_ipmp != NULL)
+		qif = qif->qf_ipmp;
+
 	READ_ENTER(&ph->ph_lock);
 
-	pfh = pfil_hook_get(flags, ph);
+	pfh = pfil_hook_get(flags & PFIL_INOUT, ph);
 	err = 0;
 
 	/*LINTED: E_CONSTANT_CONDITION*/
@@ -830,10 +783,20 @@ forced_copy:
 			err = (*pfh->pfil_func)(ip, iphlen, qif, out, qpi, mp);
 			if (err || !*mp)
 				break;
+			/*
+			 * fr_pullup may have allocated a new buffer.
+			 */
 			ip = qpi->qpi_data;
 		}
 	RW_EXIT(&ph->ph_lock);
 
+	/*
+	 * Functions called via pfil_func should only return values >= 0, so
+	 * convert any that are < 0 to be > 0 and preserve the absolute value.
+	 */
+	if (err < 0)
+		err = -err;
+
 	/*
 	 * If we still have a STREAMS message after calling the filtering
 	 * hooks, return the byte order of the fields changed above on
@@ -842,21 +805,13 @@ forced_copy:
 	 * them in some way.
 	 */
 #ifndef sparc
-	if (*mp != NULL) {
-# if SOLARIS2 >= 8
-		if (sap == IP6_DL_SAP) {
-			ip6 = (ip6_t *)ip;
-			__iplen = (u_short)ip6->ip6_plen;
-			ip6->ip6_plen = htons(__iplen);
-		} else {
-# endif
+	if ((err == 0) && (*mp != NULL)) {
+		if (sap == ETHERTYPE_IP) {
 			__iplen = (u_short)ip->ip_len;
 			__ipoff = (u_short)ip->ip_off;
 			ip->ip_len = htons(__iplen);
 			ip->ip_off = htons(__ipoff);
-# if SOLARIS2 >= 8
 		}
-# endif
 	}
 #endif 
 	return err;
@@ -1071,7 +1026,7 @@ static int pfil_property_update(dev_info_t *dip)
 }
 
 
-#if SOLARIS2 < 9
+#if SOLARIS2 == 8
 int miocpullup(mblk_t *m, size_t len)
 {
 	if (m->b_cont == NULL)
