@@ -17,6 +17,7 @@ struct uio;
 #include <sys/dlpi.h>
 #include <sys/cmn_err.h>
 #ifdef sun
+# include <sys/atomic.h>
 # include <sys/sockio.h>
 # include <sys/ksynch.h>
 # include <sys/strsubr.h>
@@ -202,7 +203,6 @@ static void pfil_printioctl(mblk_t *mp)
 /* Function:    pfilbind                                                    */
 /* Returns:     int  - 0 == success, else error                             */
 /* Parameters:  q(I) - pointer to queue                                     */
-/* Write Lock:  pfil_rw                                                     */
 /*                                                                          */
 /* Check to see if a queue (or the otherside of it) is missing a qif_t      */
 /* structure.  If neither have one then allocate a new one, else copy the   */
@@ -246,7 +246,6 @@ void pfilwput_ioctl(queue_t *q, mblk_t *mp)
 #endif
 
 	if (iocp->ioc_cr && pfil_drv_priv(iocp->ioc_cr) != 0 ) {
-printf("pfil_drv_priv fialed\n");
 		putnext(q, mp);
 		return;
 	}
@@ -315,9 +314,7 @@ printf("pfil_drv_priv fialed\n");
 
 #ifdef	SIOCGTUNPARAM
 	case SIOCGTUNPARAM :
-		WRITE_ENTER(&pfil_rw);
 		qif_attach(q);
-		RW_EXIT(&pfil_rw);
 		break;
 #endif
 
